# python 3.6
# -*- coding: utf-8 -*-
"""

The data module contains tools for preprocessing data. It allows users to merge timeseries, compute
daily and monthly summary statistics, and get seasonal periods of a time series.

"""
from __future__ import division
from hydrostats.metrics import HydrostatsError
import pandas as pd
from numpy import inf, nan
from math import modf

__all__ = ['merge_data', 'daily_average', 'daily_std_error', 'daily_std_dev', 'monthly_average',
           'monthly_std_error', 'monthly_std_dev', 'remove_nan_df', 'seasonal_period']


def merge_data(sim_fpath=None, obs_fpath=None, sim_df=None, obs_df=None, interpolate=None,
               column_names=['Simulated', 'Observed'], simulated_tz=None, observed_tz=None, interp_type='pchip'):
    """Takes two csv files or two pandas dataframes that have been formatted with 1 row as a header with date in the
    first column and streamflow values in the second column and combines them into a pandas dataframe with datetime type
     for the dates and float type for the streamflow value. Please note that the only acceptable time deltas are 15min,
    30min, 45min, and any number of hours in between.

        There are three scenarios to consider when merging your data.

        The first scenario is that the timezones and the spacing of the time series matches (eg. 1 Day). In this case,
        you will want to leave the simulated_tz, observed_tz, and interpolate arguments empty, and the function will
        simply join the two csv's into a dataframe.

        The second scenario is that you have two time series with matching time zones but not matching spacing. In this
        case you will want to leave the simulated_tz and observed_tz empty, and use the interpolate argument to tell the
        function which time series you would like to interpolate to match the other time series.

        The third scenario is that you have two time series with different time zones and possibly different spacings.
        In this case you will want to fill in the simulated_tz, observed_tz, and interpolate arguments. This will then
        take timezones into account when interpolating the selected time series.
        """
    # Reading the data into dataframes if from file
    if sim_fpath is not None and obs_fpath is not None:
        # Importing data into a data-frame
        sim_df = pd.read_csv(sim_fpath, delimiter=",", header=None, names=[column_names[0]],
                             index_col=0, infer_datetime_format=True, skiprows=1)
        obs_df = pd.read_csv(obs_fpath, delimiter=",", header=None, names=[column_names[1]],
                             index_col=0, infer_datetime_format=True, skiprows=1)
        # Converting the index to datetime type
        obs_df.index = pd.to_datetime(obs_df.index, infer_datetime_format=True, errors='coerce')
        sim_df.index = pd.to_datetime(sim_df.index, infer_datetime_format=True, errors='coerce')

    elif sim_df is not None and obs_df is not None:
        # Overriding the column names to match the column name input
        sim_df.columns = sim_df.columns.astype(str)
        sim_df.columns.values[0] = column_names[0]
        obs_df.columns = obs_df.columns.astype(str)
        obs_df.columns.values[0] = column_names[1]

        # Checking to make sure that both dataframes have datetime indices if they are not read from file.
        if sim_df.index.dtype == "datetime64[ns]" and obs_df.index.dtype == "datetime64[ns]":
            pass
        else:
            obs_df.index = pd.to_datetime(obs_df.index, infer_datetime_format=True, errors='coerce')
            sim_df.index = pd.to_datetime(sim_df.index, infer_datetime_format=True, errors='coerce')

    else:
        raise HydrostatsError('either sim_fpath and obs_fpath or sim_df and obs_df are required inputs.')

    # Checking to see if the necessary arguments in the function are fulfilled
    if simulated_tz is None and observed_tz is not None:

        raise HydrostatsError('Either Both Timezones are required or neither')

    elif simulated_tz is not None and observed_tz is None:

        raise HydrostatsError('Either Both Timezones are required or neither')

    elif simulated_tz is not None and observed_tz is not None and interpolate is None:

        raise HydrostatsError("You must specify whether to interpolate the 'simulated' or 'observed' data.")

    elif simulated_tz is None and observed_tz is None and interpolate is None:
        """Scenario 1"""
        # Merging and joining the two dataframes
        return pd.DataFrame.join(sim_df, obs_df).dropna()

    elif simulated_tz is None and observed_tz is None and interpolate is not None:
        """Scenario 2"""

        if interpolate == 'simulated':
            simulated_index_interpolate = pd.date_range(sim_df.index[0], sim_df.index[-1],
                                                        freq='15min', tz=simulated_tz)
            # Reindexing and interpolating the dataframe to match the observed data
            sim_df = sim_df.reindex(simulated_index_interpolate).interpolate(interp_type)

        elif interpolate == 'observed':

            # Condition for a half hour time delta
            # Making a new index of quarter hour time spacing for interpolation
            observed_index_interpolate = pd.date_range(obs_df.index[0], obs_df.index[-1],
                                                       freq='15min', tz=observed_tz)
            # Reindexing and interpolating the dataframe to match the observed data
            obs_df = obs_df.reindex(observed_index_interpolate).interpolate(interp_type)

        else:
            raise HydrostatsError("The interpolate argument must be either 'simulated' or 'observed'.")

        return pd.DataFrame.join(sim_df, obs_df).dropna()

    elif simulated_tz is not None and observed_tz is not None and interpolate is not None:
        """Scenario 3"""

        # Finding the frequency of the timeseries for observed and simulated
        td_simulated = (sim_df.index[1] - sim_df.index[0]).days + \
                       ((sim_df.index[1] - sim_df.index[0]).seconds / 3600) / 24
        td_observed = (obs_df.index[1] - obs_df.index[0]).days + \
                      ((obs_df.index[1] - obs_df.index[0]).seconds / 3600) / 24

        # converting the time delta to a tuple with days and hours
        td_tuple_simulated = modf(td_simulated)
        td_tuple_observed = modf(td_observed)

        # Converting the time delta to a frequency
        freq_simulated = str(td_tuple_simulated[1]) + 'D' + str(td_tuple_simulated[0] * 24) + 'H'
        freq_observed = str(td_tuple_observed[1]) + 'D' + str(td_tuple_observed[0] * 24) + 'H'

        # Making a new index for reindexing the time series
        simulated_df_new_index = pd.date_range(sim_df.index[0], sim_df.index[-1],
                                               freq=freq_simulated, tz=simulated_tz)
        observed_df_new_index = pd.date_range(obs_df.index[0], obs_df.index[-1],
                                              freq=freq_observed, tz=observed_tz)

        # Changing the time series index to reflect the changes in the timezones
        sim_df.index = simulated_df_new_index
        obs_df.index = observed_df_new_index

        if interpolate == 'simulated':
            # Making a new index of quarter hour time spacing for interpolation
            simulated_index_interpolate = pd.date_range(sim_df.index[0], sim_df.index[-1],
                                                        freq='15min', tz=simulated_tz)
            # Reindexing and interpolating the dataframe to match the observed data
            sim_df = sim_df.reindex(simulated_index_interpolate).interpolate(interp_type)

        elif interpolate == 'observed':
            # Making a new index of quarter hour time spacing for interpolation
            observed_index_interpolate = pd.date_range(obs_df.index[0], obs_df.index[-1],
                                                       freq='15min', tz=observed_tz)
            # Reindexing and interpolating the dataframe to match the observed data
            obs_df = obs_df.reindex(observed_index_interpolate).interpolate(interp_type)

        else:
            raise HydrostatsError("You must specify the interpolation argument to be either 'simulated' or "
                                        "'observed'.")

        return pd.DataFrame.join(sim_df, obs_df).dropna()


def daily_average(merged_data):
    """Takes a dataframe with an index of datetime type and both recorded and predicted streamflow values and
        calculates the daily average streamflow over the course of the data. Returns a dataframe with the date
        format as MM/DD in the index and the two columns of the data averaged"""
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m/%d"))
    return a.mean()


def daily_std_error(merged_data):
    """Takes a dataframe with an index of datetime type and both recorded and predicted streamflow values and
        calculates the daily standard error of the streamflow over the course of the data. Returns a dataframe with
        the date format as MM/DD in the index and the two columns of the data's standard error"""
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m/%d"))
    return a.sem()


def daily_std_dev(merged_data):
    """Takes a dataframe with an index of datetime type and both recorded and predicted streamflow values and
        calculates the daily standard deviation of the streamflow over the course of the data. Returns a dataframe with
        the date format as MM/DD in the index and the two columns of the data's standard deviation"""
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m/%d"))
    return a.std()


def monthly_average(merged_data):
    """Takes a dataframe with an index of datetime type and both recorded and predicted streamflow values and
        calculates the monthly average streamflow over the course of the data. Returns a dataframe with the date
        format as MM in the index and the two columns of the data averaged"""
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m"))
    return a.mean()


def monthly_std_error(merged_data):
    """Takes a dataframe with an index of datetime type and both recorded and predicted streamflow values and
        calculates the monthly standard error of the streamflow over the course of the data. Returns a dataframe
        with the date format as MM in the index and the two columns of the data's standard error"""
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m"))
    return a.sem()


def monthly_std_dev(merged_data):
    """Takes a dataframe with an index of datetime type and both recorded and predicted streamflow values and
        calculates the monthly standard deviation of the streamflow over the course of the data.
        Returns a dataframe with the date format as MM in the index and the two columns of the data's standard
        deviation"""
    # Calculating the daily average from the database
    a = merged_data.groupby(merged_data.index.strftime("%m"))
    return a.std()


def remove_nan_df(merged_dataframe):
    """Drops rows with Nan, zero, negative, and inf values from a pandas dataframe"""
    # Drops Zeros and negatives
    merged_dataframe = merged_dataframe.loc[~(merged_dataframe <= 0).any(axis=1)]
    # Replaces infinites with nans
    merged_dataframe = merged_dataframe.replace([inf, -inf], nan)
    # Drops Nan
    merged_dataframe = merged_dataframe.dropna()
    return merged_dataframe


def seasonal_period(merged_dataframe, daily_period, time_range=None, numpy=False):
    """Returns the seasonal period specified for the time series, between the specified time range. Can return either a
    pandas DataFrame or two numpy arrays"""
    # Making a copy to avoid changing the original df
    merged_df_copy = merged_dataframe.copy()

    if time_range:
        # Setting the time range
        merged_df_copy = merged_df_copy.loc[time_range[0]: time_range[1]]

    # Setting a placeholder for the datetime string values
    merged_df_copy.insert(loc=0, column='placeholder', value=merged_df_copy.index.strftime('%m-%d'))

    # getting the start and end of the seasonal period
    start = daily_period[0]
    end = daily_period[1]

    # Getting the seasonal period
    if start < end:
        merged_df_copy = merged_df_copy.loc[(merged_df_copy['placeholder'] >= start) &
                                                (merged_df_copy['placeholder'] <= end)]
    else:
        merged_df_copy = merged_df_copy.loc[(merged_df_copy['placeholder'] >= start) |
                                                (merged_df_copy['placeholder'] <= end)]
    # Dropping the placeholder
    merged_df_copy = merged_df_copy.drop(columns=['placeholder'])

    if numpy:
        return merged_df_copy.iloc[:, 0].values, merged_df_copy.iloc[:, 1].values
    else:
        return merged_df_copy



# python 3.6
# -*- coding: utf-8 -*-
"""

The analyze module contains functions that perform a more complex analysis of simulated and observed
time series data. It allows users to make tables with metrics that they choose as well as different
date ranges. It also allows users to run a time lag analysis of two time series.

"""
from __future__ import division
from hydrostats.metrics import list_of_metrics, metric_names, metric_abbr, remove_values, HydrostatsError
from hydrostats.data import seasonal_period
import calendar
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

__all__ = ['make_table', 'time_lag']


def make_table(merged_dataframe, metrics, seasonal_periods=None, mase_m=1, dmod_j=1,
               nse_mod_j=1, h6_mhe_k=1, h6_ahe_k=1, h6_rmshe_k=1, d1_p_obs_bar_p=None,
               lm_x_obs_bar_p=None, replace_nan=None, replace_inf=None, remove_neg=False,
               remove_zero=False, to_csv=None, to_excel=None, location=None):
    """

    Create a table of user selected metrics with optional seasonal analysis.

    **Notes:** Creates a table with metrics as specified by the user. Seasonal periods can also be
    specified in order to compare different seasons and how well the simulated data matches the
    observed data. Has options to save the table to either a csv or an excel workbook. Also has
    an option to add a column for the location of the data.

    **Parameters**

    merged_dataframe (required): *DataFrame*
    > A pandas dataframe input that has two columns of predicted data and observed data with a
    datetime index.

    metrics (required): *list*
     > A list of all the metrics that the user wants to calculate. The metrics abbreviations must
     be used and can be found in the table of all the metrics in the documentation.

    seasonal_periods: "2D List"
    > If give, specifies the seasonal periods that the user wants to analyze (e.g. [['06-01',
    '06-30'], ['08-12', '11-23']] would analyze the dates from June 1st to June 30th and also
    August 8th to November 23). Note that the entire time series is analyzed with the selected
    metrics by default.

    mase_m: *integer*
    > Optional parameter for the mean absolute scaled error (MASE) metric.

    dmod_j: *integer or float*
    > Optional parameter for the modified index of agreement (dmod) metric.

    nse_mod_j: *integer or float*
    > Optional parameter for the modified Nash Sutcliffe (nse_mod) metric.

    h6_mhe_k: *integer or float*
    > Optional parameter for the H6 (MHE) metric.

    h6_ahe_k: *integer or float*
    > Optional parameter for the H6 (AHE) metric

    h6_rmshe_k: *integer or float*
    > Optional parameter for the H6 (RMSHE) metric

    d1_p_obs_bar_p: *float*
    > Optional parameter fot the Legate McCabe Index of Agreement (d1_p).

    lm_x_obs_bar_p: *float*
    > Optional parameter for the Lagate McCabe Efficiency Index (lm_index).

    replace_nan: *float, optional*
    > If given, indicates which value to replace NaN values with in the two arrays. If None, when
    a NaN value is found at the i-th position in the observed OR simulated array, the i-th value
    of the observed and simulated array are removed before the computation.

    replace_inf *float, optional*
    > If given, indicates which value to replace Inf values with in the two arrays. If None, when
    an inf value is found at the i-th position in the observed OR simulated array, the i-th value
    of the observed and simulated array are removed before the computation.

    remove_neg *boolean, optional*
    > If true, negative numbers will be removed from the observed and simulated arrays.

    remove_zero *boolean, optional*
    > If true, zeros will be removed from the observed and simulated arrays.

    to_csv: *string*
    > Filepath and file name of the csv that is written (e.g. r'/path/to/output_dir/file.csv').

    to_excel: *string*
    > Filepath and file name of the excel workbook that is written
    (e.g. r'/path/to/output_dir/file.xlsx').

    location: *string*
    > The name of the location that will be created as a column in the table that is created.
    Useful for creating a large table with different datasets.

    **Example**

    ```python

    import hydrostats as hs
    import hydrostats.data as hd

    # Defining the URLs of the datasets
    sfpt_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/sfpt_data/magdalena-calamar_interim_data.csv'
    glofas_url = r'https://github.com/waderoberts123/Hydrostats/raw/master/Sample_data/GLOFAS_Data/magdalena-calamar_ECMWF_data.csv'

    # Merging the data
    merged_df = hd.merge_data(sfpt_url, glofas_url, column_names=['SFPT', 'GLOFAS'])

    # Printing the table
    print(hs.make_table(merged_dataframe=merged_df,
                        metrics=['MAE', 'r2', 'ACC', 'NSE', 'SA'],
                        seasonal_periods=[['01-01', '03-31'], ['04-01', '06-30'], ['07-01', '09-30'],
                                          ['10-01', '12-31']],
                        remove_neg=True, remove_zero=True, location='Magdalena'))

    # Writing the table to a csv
    hs.make_table(merged_dataframe=merged_df,
                        metrics=['MAE', 'r2', 'ACC', 'NSE', 'SA'],
                        seasonal_periods=[['01-01', '03-31'], ['04-01', '06-30'], ['07-01', '09-30'],
                                          ['10-01', '12-31']],
                        remove_neg=True, remove_zero=True, location='Magdalena',
                        to_csv='magdalena_table.csv')
    ```

    """

    # Creating a list for all of the metrics for all of the seasons
    complete_metric_list = []

    # Creating an index list
    index_array = ['Full Time Series']
    if seasonal_periods is not None:
        seasonal_periods_names = []
        for i in seasonal_periods:
            month_1 = calendar.month_name[int(i[0][:2])]
            month_2 = calendar.month_name[int(i[1][:2])]
            name = month_1 + i[0][2:] + ':' + month_2 + i[1][2:]
            seasonal_periods_names.append(name)
        index_array.extend(seasonal_periods_names)

    # Creating arrays for sim and obs with all the values if a merged dataframe is given
    sim_array = merged_dataframe.iloc[:, 0].values
    obs_array = merged_dataframe.iloc[:, 1].values

    # Getting a list of the full time series
    full_time_series_list = list_of_metrics(metrics=metrics, abbr=True, sim_array=sim_array, obs_array=obs_array,
                                            mase_m=mase_m, dmod_j=dmod_j, nse_mod_j=nse_mod_j, h6_mhe_k=h6_mhe_k,
                                            h6_ahe_k=h6_ahe_k, h6_rmshe_k=h6_rmshe_k, d1_p_obs_bar_p=d1_p_obs_bar_p,
                                            lm_x_obs_bar_p=lm_x_obs_bar_p, replace_nan=replace_nan,
                                            replace_inf=replace_inf, remove_neg=remove_neg, remove_zero=remove_zero)

    # Appending the full time series list to the entire list:
    complete_metric_list.append(full_time_series_list)

    # Calculating metrics for the seasonal periods
    if seasonal_periods is not None:
        for time in seasonal_periods:
            temp_df = seasonal_period(merged_dataframe, time)
            sim_array = temp_df.iloc[:, 0].values
            obs_array = temp_df.iloc[:, 1].values

            seasonal_metric_list = list_of_metrics(metrics=metrics, sim_array=sim_array, abbr=True, obs_array=obs_array,
                                                   mase_m=mase_m, dmod_j=dmod_j, nse_mod_j=nse_mod_j,
                                                   h6_mhe_k=h6_mhe_k, h6_ahe_k=h6_ahe_k, h6_rmshe_k=h6_rmshe_k,
                                                   d1_p_obs_bar_p=d1_p_obs_bar_p, lm_x_obs_bar_p=lm_x_obs_bar_p,
                                                   replace_nan=replace_nan, replace_inf=replace_inf,
                                                   remove_neg=remove_neg, remove_zero=remove_zero)

            complete_metric_list.append(seasonal_metric_list)

    table_df_final = pd.DataFrame(complete_metric_list, index=index_array, columns=metrics)

    if location is not None:
        col_values = [location for i in range(table_df_final.shape[0])]
        table_df_final.insert(loc=0, column='Location', value=np.array(col_values))

    if to_csv is None and to_excel is None:
        return table_df_final

    elif to_csv is None and to_excel is not None:
        table_df_final.to_excel(to_excel, index_label='Datetime')

    elif to_csv is not None and to_excel is None:
        table_df_final.to_csv(to_csv, index_label='Datetime')

    else:
        table_df_final.to_excel(to_excel, index_label='Datetime')
        table_df_final.to_csv(to_csv, index_label='Datetime')


def time_lag(merged_dataframe, metrics, interp_freq='6H', interp_type='pchip',
             shift_range=[-30, 30], mase_m=1, dmod_j=1, nse_mod_j=1, h6_mhe_k=1,
             h6_ahe_k=1, h6_rmshe_k=1, d1_p_obs_bar_p=None, lm_x_obs_bar_p=None, replace_nan=None,
             replace_inf=None, remove_neg=False, remove_zero=False,
             plot_title='Metric Values as Different Lags', ylabel='Metric Value',
             xlabel='Number of Lags', save_fig=None, figsize=(10, 6), station=None, to_csv=None,
             to_excel=None):
    """

    Check metric values between simulated and observed data at different time lags.



    """

    metrics_list = metric_names
    abbreviations = metric_abbr

    abbr_indices = []
    for i in metrics:
        abbr_indices.append(metrics_list.index(i))

    abbr_list = []
    for i in abbr_indices:
        abbr_list.append(abbreviations[i])

    # Making a new time index to be able to interpolate the time series to the required input
    new_index = pd.date_range(merged_dataframe.index[0], merged_dataframe.index[-1], freq=interp_freq)

    # Reindexing the dataframe and interpolating it
    try:
        merged_dataframe = merged_dataframe.reindex(new_index)
        merged_dataframe = merged_dataframe.interpolate(interp_type)
    except ValueError:
        raise HydrostatsError('ValueError Raised while interpolating, you may want to check for duplicates in'
                              ' your dates.')

    # Making arrays to compare the metric value at different time steps
    sim_array = merged_dataframe.iloc[:, 0].values
    obs_array = merged_dataframe.iloc[:, 1].values

    sim_array, obs_array = remove_values(sim_array, obs_array, replace_nan=replace_nan, replace_inf=replace_inf,
                                         remove_zero=remove_zero, remove_neg=remove_neg)

    # Creating a list to append the values of shift to
    shift_list = []

    # Creating a list of all the time shifts specified by the user
    lag_array = np.arange(shift_range[0], shift_range[1] + 1)

    # Looping through the list of lags and appending the metric value to the shift list
    for i in lag_array:
        sim_array_temp = np.roll(sim_array, i)

        lag_metrics = list_of_metrics(metrics=metrics, sim_array=sim_array_temp, obs_array=obs_array, mase_m=mase_m,
                                      dmod_j=dmod_j, nse_mod_j=nse_mod_j, h6_mhe_k=h6_mhe_k,
                                      h6_ahe_k=h6_ahe_k, h6_rmshe_k=h6_rmshe_k, d1_p_obs_bar_p=d1_p_obs_bar_p,
                                      lm_x_obs_bar_p=lm_x_obs_bar_p, replace_nan=replace_nan, replace_inf=replace_inf,
                                      remove_neg=remove_neg, remove_zero=remove_zero)
        shift_list.append(lag_metrics)

    final_array = np.array(shift_list)

    plt.figure(figsize=figsize)

    for i, abbr in enumerate(abbr_list):
        shift_list_temp = final_array[:, i]
        plt.plot(lag_array, shift_list_temp, label=abbr, alpha=0.7)

    plt.title(plot_title, fontsize=18)
    plt.xlabel(xlabel, fontsize=14)
    plt.ylabel(ylabel, fontsize=14)
    plt.xticks(fontsize=12)
    plt.yticks(fontsize=12)
    plt.legend()
    if save_fig is None:
        plt.show()
    else:
        plt.savefig(save_fig)
        plt.close()

    max_lag_array = np.max(final_array, 0)
    max_lag_indices = np.argmax(final_array, 0)
    max_lag_locations = lag_array[max_lag_indices]
    min_lag_array = np.min(final_array, 0)
    min_lag_indices = np.argmin(final_array, 0)
    min_lag_locations = lag_array[min_lag_indices]

    data = np.column_stack((max_lag_array, max_lag_locations, min_lag_array, min_lag_locations))

    final_df = pd.DataFrame(data=data, index=metrics, columns=["Max", "Max Lag Number", "Min", "Min Lag Number"])

    if station is not None:
        col_values = [station for i in range(final_df.shape[0])]
        final_df.insert(loc=0, column='Station', value=np.array(col_values))

    if to_csv is None and to_excel is None:
        return final_df

    elif to_csv is None and to_excel is not None:
        final_df.to_excel(to_excel, index_label='Metric')

    elif to_csv is not None and to_excel is None:
        final_df.to_csv(to_csv, index_label='Metric')

    else:
        final_df.to_excel(to_excel, index_label='Metric')
        final_df.to_csv(to_csv, index_label='Metric')
